#include "libc.h"
#include "sim_types.h"
#include "sim_seq.h"
#include "sim_args.h"
#include "sim_misc.h"
#include "sim_karlin.h"
#include "sim_prot.h"

static const char rcsid[] =
"$Id: prot.c,v 1.3 1998/06/06 09:25:49 schwartz Exp schwartz $";

static const argv_scores_t EIMOSV = 
{ DEFAULT_AA_E, 0, 0, DEFAULT_AA_O, 0.0, 0 };

/* amino acid names */
const uchar pchars[] = "ARNDCQEGHILKMFPSTWYVBZX";

static const int blosum62[23][23] = {
   { 4,-1,-2,-2, 0,-1,-1, 0,-2,-1,-1,-1,-1,-2,-1, 1, 0,-3,-2, 0,-2,-1, 0 },
   {-1, 5, 0,-2,-3, 1, 0,-2, 0,-3,-2, 2,-1,-3,-2,-1,-1,-3,-2,-3,-1, 0,-1 },
   {-2, 0, 6, 1,-3, 0, 0, 0, 1,-3,-3, 0,-2,-3,-2, 1, 0,-4,-2,-3, 3, 0,-1 },
   {-2,-2, 1, 6,-3, 0, 2,-1,-1,-3,-4,-1,-3,-3,-1, 0,-1,-4,-3,-3, 4, 1,-1 },
   { 0,-3,-3,-3, 9,-3,-4,-3,-3,-1,-1,-3,-1,-2,-3,-1,-1,-2,-2,-1,-3,-3,-2 },
   {-1, 1, 0, 0,-3, 5, 2,-2, 0,-3,-2, 1, 0,-3,-1, 0,-1,-2,-1,-2, 0, 3,-1 },
   {-1, 0, 0, 2,-4, 2, 5,-2, 0,-3,-3, 1,-2,-3,-1, 0,-1,-3,-2,-2, 1, 4,-1 },
   { 0,-2, 0,-1,-3,-2,-2, 6,-2,-4,-4,-2,-3,-3,-2, 0,-2,-2,-3,-3,-1,-2,-1 },
   {-2, 0, 1,-1,-3, 0, 0,-2, 8,-3,-3,-1,-2,-1,-2,-1,-2,-2, 2,-3, 0, 0,-1 },
   {-1,-3,-3,-3,-1,-3,-3,-4,-3, 4, 2,-3, 1, 0,-3,-2,-1,-3,-1, 3,-3,-3,-1 },
   {-1,-2,-3,-4,-1,-2,-3,-4,-3, 2, 4,-2, 2, 0,-3,-2,-1,-2,-1, 1,-4,-3,-1 },
   {-1, 2, 0,-1,-3, 1, 1,-2,-1,-3,-2, 5,-1,-3,-1, 0,-1,-3,-2,-2, 0, 1,-1 },
   {-1,-1,-2,-3,-1, 0,-2,-3,-2, 1, 2,-1, 5, 0,-2,-1,-1,-1,-1, 1,-3,-1,-1 },
   {-2,-3,-3,-3,-2,-3,-3,-3,-1, 0, 0,-3, 0, 6,-4,-2,-2, 1, 3,-1,-3,-3,-1 },
   {-1,-2,-2,-1,-3,-1,-1,-2,-2,-3,-3,-1,-2,-4, 7,-1,-1,-4,-3,-2,-2,-1,-2 },
   { 1,-1, 1, 0,-1, 0, 0, 0,-1,-2,-2, 0,-1,-2,-1, 4, 1,-3,-2,-2, 0, 0, 0 },
   { 0,-1, 0,-1,-1,-1,-1,-2,-2,-1,-1,-1,-1,-2,-1, 1, 5,-2,-2, 0,-1,-1, 0 },
   {-3,-3,-4,-4,-2,-2,-3,-2,-2,-3,-2,-3,-1, 1,-4,-3,-2,11, 2,-3,-4,-3,-2 },
   {-2,-2,-2,-3,-2,-1,-2,-3, 2,-1,-1,-2,-1, 3,-3,-2,-2, 2, 7,-1,-3,-2,-1 },
   { 0,-3,-3,-3,-1,-2,-2,-3,-3, 3, 1,-2, 1,-1,-2,-2, 0,-3,-1, 4,-3,-2,-1 },
   {-2,-1, 3, 4,-3, 0, 1,-1, 0,-3,-4, 0,-3,-3,-2, 0,-1,-4,-3,-3, 4, 1,-1 },
   {-1, 0, 0, 1,-3, 3, 4,-2, 0,-3,-3, 1,-1,-3,-1, 0,-1,-3,-2,-2, 1, 4,-1 },
   { 0,-1,-1,-1,-2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-2, 0, 0,-2,-1,-1,-1,-1,-1 },
 };

/* protein_scores - set scoring matrix for proteins */
void protein_scores(argv_scores_t *ps, ss_t ss)
{
        unsigned int i, j;

	argv_scores(ps, &EIMOSV);
	for (i = 0; i < CLEN(pchars); ++i)
		for (j = 0; j < CLEN(pchars); ++j)
			ss[pchars[i]][pchars[j]] = blosum62[i][j];
	for (i = 0; i < CLEN(pchars); ++i)
		ss[pchars[i]]['X'] =  ss['X'][pchars[i]] = -1000;
}

/* protein_thresh ------- determine cutoff score according to Karlin-Altschul */
int protein_thresh(SEQ *seq1, SEQ *seq2, double S)
{
    int achars[NACHARS];
    int a[CLEN(pchars)], b[CLEN(pchars)], maximum, minimum, k;
    double p[1000], *q, N, x, param_K, param_lambda;
    int i, j;
    
    for (i = 0; pchars[i]; i++) {
	achars[pchars[i]] = i;
    }
    for (i = 0; i < (int)CLEN(pchars); i++) a[i] = b[i] = 0;
    for (i = 0; i < SEQ_LEN(seq1); i++) {
	a[achars[SEQ_AT(seq1,i)]]++;
    }
    for (i = 0; i < SEQ_LEN(seq2); i++) {
	b[achars[SEQ_AT(seq2,i)]]++;
    }
    for (i = 0; i < 1000; i++) p[i] = 0.0;
    for (maximum = minimum = i = 0; i < 23; i++)
	for (j = 0; j < (int)CLEN(pchars); j++) 
	    if (maximum < (k=blosum62[i][j])) maximum = k;
	    else if (minimum > k) minimum = k;
    q = p - minimum; 
    for (i = 0; i < (int)CLEN(pchars); i++) {
	for (j = 0; j < (int)CLEN(pchars); j++) {
	    k = blosum62[i][j];
	    q[k] += (double)  a[i] * (double) b[j];
	}
    }
    if (SEQ_SAME(seq1,seq2)) {
	N = (double) (SEQ_LEN(seq1)-1) * (double)SEQ_LEN(seq2)/2;
	for (i = 0; i < (int)CLEN(pchars); i++) {
	    k = blosum62[i][i];
	    q[k] -= (double) a[i];
	}
	for (i = minimum; i <= maximum; i++) 
	    q[i] /= 2.0;
    } else {
	N = (double) SEQ_LEN(seq1) * (double) SEQ_LEN(seq2);
    }
    for (i = minimum; i <= maximum; i++)
	q[i] /= N;

    if (karlin(minimum, maximum, p, &param_lambda, &param_K) == 0)
	fatal("computation of significance levels failed");
    
    x = log(S)/(-param_K*N);
    i =  2*log(sqrt(x))/(-param_lambda);
    return i;
}

int is_pchar(int ch)
{
	return !!strchr((const char*)pchars, ch);
}
