#include "libc.h"
#include "sim_types.h"
#include "sim_misc.h"
#include "sim_seq.h"

static const char rcsid[] =
"$Id: seq.c,v 1.5 1998/06/06 22:35:32 schwartz Exp schwartz $";

static uchar ckgetc(SEQ *s);

static int parse_fname(const char* arg, char **fname, int *from, int *len)
{
        char *p = 0;

        *fname = copy_string(arg);
        if ((p = strchr(*fname, '{')) || 
	    (p = strchr(*fname, '[')) ||
            (p = strchr(*fname, '('))) {
	        int to;
		if (sscanf(p+1, "%d,%d", from, &to) != 2)
			return -1;
		if (*from <= 0 || *from > to)
			return -1;
		*p = '\0';
		*len = to - *from + 1;
		return 1;
	} else {
		*from = 1;
		*len = -1;
		return 0;
	}
}

SEQ* seq_open(const char *fname, const char *mode, int flags)
{
	SEQ *s = ckalloc(sizeof(SEQ));

	mode = 0;
	flags = 0;

	switch (parse_fname(fname, &(s->fname), &(s->from), &(s->slen))) {
	case -1:
		fatalf("improper positions specification: %s", fname);
		break;
	case 1:
		s->flags |= SIM_FLAG_SUBRANGE;
		break;
	default:
		break;
	}

	s->fp = ckopen(s->fname, "r");
	s->count = 0;
	return s;
}

void seq_close(SEQ *s)
{
	fclose(s->fp);
	free(s->fname);
	free(s->header);
	free(s->seq);
	memset(s, 0, sizeof(SEQ));
	free(s);
}

int seq_read(SEQ *seq)
{
	long fpos;
	int i, n;
	int b, c;
	uchar *p;
	char *q;

	if (feof(seq->fp))
		return 0;

	if (seq->count > 0) { 
		if (seq->flags & SIM_FLAG_SUBRANGE) {
			return -1;
		} else {
			seq->from = 1;
			seq->slen = -1;
		}
	}

	if (seq->header) ZFREE(seq->header);
	if (seq->seq) ZFREE(seq->seq);

	/* step over header */
	fpos = ckftell(seq->fp);
	n = 0;
	b = '\n';
	c = getc(seq->fp);

	while ((c != EOF) && !(isalpha(c) && (b=='\n'))) {
		++n;
		b = c;
		c = getc(seq->fp);
	}
	if (c != EOF)
		ungetc(c, seq->fp);
	else
		return 0;

	/* go back and copy header */
	seq->hlen = n;
	seq->header = ckalloc((size_t)n+1);
	ckfseek(seq->fp, fpos, SEEK_SET);
	for (q = seq->header; n--; ++q)
		*q = getc(seq->fp);
	*q = 0;


	/* step over seq */
	fpos = ckftell(seq->fp);
	n = 0;
	b = '\n';
	c = getc(seq->fp);
	while ((c != EOF) && !(!isalpha(c) && (b=='\n'))) {
		if (!isspace(c)) ++n;
		b = c;
		c = getc(seq->fp);
	}
	if (c != EOF) ungetc(c, seq->fp);

	/* check conformance */
	if (seq->slen == -1)
		seq->slen = n;
	if (seq->slen > n) {
		char buf[1024];
		sprintf(buf, "sequence %d at %ld is too long (%d > %d)",
			seq->count, fpos, seq->slen, n);
		fatal(buf);
	}

	/* go back and copy seq */
	seq->seq = ckalloc((size_t)n+1);
	ckfseek(seq->fp, fpos, SEEK_SET);
	for (i = seq->from-1; i; --i)
		c = getc(seq->fp);
	p = seq->seq;
	while (n) {
		c = ckgetc(seq);
		if (!isspace(c)) {
			*p++ = c;
			--n;
		}
	}
	c = getc(seq->fp);
	*p = 0;

	seq->count++;
	return 1;
}

/* ckgetc - get a character, which must be a letter (made upper case) or '\n' */
static uchar ckgetc(SEQ *s)
{
	int c;

	if ((c = getc(s->fp)) == EOF)
		fatalf("unexpected EOF in \"%s\".", s->fname);
	if (islower(c))
		c = toupper(c);
	if (c != '\n' && !isalpha(c)) {
		(void)fprintf(stderr, "illegal character '%c' in file %s\n",
		   	      c, s->fname);
		exit(1);
	}
	return (uchar)c;
}

